r'''
Copy with a memory.

(c) Tom de Geus, 2021, MIT
'''

from ._version import *
from . import checksum
from . import convert
from . import detail
from . import path
from . import relpath
from . import rich
from . import rsync
from . import scp
from . import ssh
from . import yaml









'''shelephant_rm
    Remove files listed in a (field of a) YAML-file.
    The filenames are assumed either absolute, or relative to the input YAML-file.

:usage:

    shelephant_rm [options]

    shelephant_rm [options] <input.yaml>

:argument:

    YAML-file with filenames. Default: shelephant_dump.yaml

:options:

    -k, --key=N
        Path in the YAML-file, separated by "/". [default: /]

    -f, --force
        Remove without prompt.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import click
import os

from .. import relpath
from .. import version
from .. import yaml


def main():

    try:

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-k', '--key', required=False, default='/')
        parser.add_argument('-f', '--force', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('input', nargs='?', default='shelephant_dump.yaml')
        args = parser.parse_args()

        source = args.input
        key = list(filter(None, args.key.split('/')))
        files = yaml.read_item(source, key)
        prefix = os.path.dirname(source)
        files = relpath.add_prefix(prefix, files)

        if len(files) == 0:
            return 0

        for file in files:
            print('rm {0:s}'.format(file))

        if not args.force:
            if not click.confirm('Proceed?'):
                return 1

        for file in files:
            os.remove(file)

    except Exception as e:

        print(e)
        return 1


if __name__ == '__main__':

    main()



'''shelephant_checksum
    Get checksum of files listed in a (field of a) YAML-file.
    The filenames are assumed either absolute, or relative to the input YAML-file.

:usage:

    shelephant_checksum [options] <input.yaml>

:argument:

    YAML-file with file-paths. Default: shelephant_dump.yaml

:options:

    -o, --output=N
        Output YAML-file. [default: shelephant_checksum.yaml]

    -k, --key=N
        Path in the YAML-file, separated by "/". [default: /]

    -l, --local=N
        Add local 'host' information to use precomputed checksums.

    -f, --force
        Overwrite output file without prompt.

    -q, --quiet
        Do not print progress.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import os

from .. import checksum
from .. import relpath
from .. import version
from .. import yaml


def main():

    try:

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-o', '--output', required=False, default='shelephant_checksum.yaml')
        parser.add_argument('-k', '--key', required=False, default='/')
        parser.add_argument('-l', '--local', required=False)
        parser.add_argument('-f', '--force', required=False, action='store_true')
        parser.add_argument('-q', '--quiet', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('input', nargs='?', default='shelephant_dump.yaml')
        args = parser.parse_args()

        source = args.input
        key = list(filter(None, args.key.split('/')))
        files = yaml.read_item(source, key)
        prefix = os.path.dirname(source)
        files = relpath.add_prefix(prefix, files)
        data = checksum.get(files, args.local, hybrid=True, progress=not args.quiet)
        yaml.dump(args.output, data, args.force)

    except Exception as e:

        print(e)
        return 1


if __name__ == '__main__':

    main()



'''shelephant_parse
    Parse a YAML-file, and print to screen.

:usage:

    shelephant_parse <file.yaml>

:argument:

    File path.

:options:

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse

from .. import version
from .. import yaml


def main():

    try:

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('file')
        args = parser.parse_args()

        data = yaml.read(args.file)
        yaml.view(data)

    except Exception as e:

        print(e)
        return 1


if __name__ == '__main__':

    main()



r'''
Convert nested ``list`` or ``dict``.

(c) Tom de Geus, 2021, MIT
'''

from collections import defaultdict
import collections.abc

def _flatten_detail(data):
    r'''
Detail of :py:fun:`Flatten`.
Not part of public API.

See https://stackoverflow.com/a/17485785/2646505
    '''

    for item in data:
        if isinstance(item, collections.abc.Iterable) and not isinstance(item, str):
            for x in flatten(item):
                yield x
        else:
            yield item


def flatten(data):
    r'''
Flatten a nested list to a one dimensional list.
    '''
    return list(_flatten_detail(data))


def _squash_detail(data, parent_key='', sep='_'):
    r'''
Detail of :py:fun:`squash`.
Not part of public API.

See https://stackoverflow.com/a/6027615/2646505
    '''

    items = []

    for k, v in data.items():

        new_key = parent_key + sep + k if parent_key else k

        if isinstance(v, collections.abc.MutableMapping):
            items.extend(_squash_detail(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))

    return dict(items)


def squash(data):
    r'''
Squash a dictionary to a single list.
    '''

    return flatten(list(_squash_detail(data).values()))





r'''shelephant_hostinfo
    Collect file information from location (on a remote host).

:usage:

    shelephant_hostinfo [options]

:options:

    -o, --output=N
        Output YAML-file. [default: shelephant_hostinfo.yaml].

    -r, --host=N
        Host-name.

    -p, --prefix=N
        Directory (on host) from which to copy.

    -f, --files=[N]
        YAML-file with list of files (on remote). [default: shelephant_dump.yaml]

    -c, --checksum=[N]
        YAML-file with checksums (on remote). [default: shelephant_checksum.yaml]

    --files-key=N
        Path in the YAML-file, separated by "/". [default: /]

    --checksum-key=N
        Path in the YAML-file, separated by "/". [default: /]

    --ignore
        Skip basic checks.

    -i, --input=[N]
        Use host and prefix from existing file. [default: shelephant_hostinfo.yaml]

    --remove=N...
        Remove paths (and checksums) from existing file.

    --force
        Overwrite output file without prompt.

    --verbose
        Verbose all commands.

    -h, --help
        Show help.

    --version
        Show version.


(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import os
import tempfile
import shutil
import numpy as np

from .. import scp
from .. import version
from .. import yaml


def remove(data, rm):

    files = np.array(data['files'])
    sorter = np.argsort(files)
    files = files[sorter]
    rm = np.sort(rm)

    test = np.in1d(rm, files)
    if not np.all(np.in1d(rm, files)):
        raise IOError('One or more remove paths not found')

    idx = np.searchsorted(files, rm)
    keep = np.setdiff1d(np.arange(len(files)), sorter[idx])

    files = data['files']
    checksum = data['checksum']
    data['files'] = [files[i] for i in keep]
    data['checksum'] = [checksum[i] for i in keep]

    return data


def main():

    try:

        # Parse command-line arguments

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-o', '--output', required=False, default='shelephant_hostinfo.yaml')
        parser.add_argument('-r', '--host', required=False, default=None)
        parser.add_argument('-p', '--prefix', required=False, default=None)
        parser.add_argument('-f', '--files', required=False, default=None, nargs='?', const='shelephant_dump.yaml')
        parser.add_argument('-c', '--checksum', required=False, default=None, nargs='?', const='shelephant_checksum.yaml')
        parser.add_argument(      '--files-key', required=False, default='/')
        parser.add_argument(      '--checksum-key', required=False, default='/')
        parser.add_argument(      '--ignore', required=False, action='store_true')
        parser.add_argument('-i', '--input', required=False, default=None, nargs='?', const='shelephant_hostinfo.yaml')
        parser.add_argument(      '--remove', required=False, nargs='*')
        parser.add_argument(      '--force', required=False, action='store_true')
        parser.add_argument(      '--verbose', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        args = parser.parse_args()

        # Separate mode: remove paths and quit

        if args.remove:
            filename = args.output
            if args.input:
                filename = args.input
            data = yaml.read(filename)
            data = remove(data, args.remove)
            yaml.dump(args.output, data, args.force)
            return 0

        # Extract basic information from command-line arguments

        data = {}

        if args.host:
            data['host'] = args.host

        if args.prefix:
            data['prefix'] = args.prefix

        if args.input:
            overwrite = yaml.read(args.input)
            for item in ['host', 'prefix']:
                if item in overwrite:
                    data[item] = overwrite[item]

        # Basic IO-checks

        if not args.ignore:
            if 'host' in data and 'prefix' not in data:
                raise IOError('Specify hostname and prefix')

        if 'host' not in data and 'prefix' not in data:
            for item in [args.files, args.checksum]:
                if item:
                    data['prefix'] = os.path.dirname(item)
                    break

        # Create temporary file to download to

        if 'host' in data:
            temp_dir = tempfile.mkdtemp()
            temp_file = os.path.join(temp_dir, 'shelephant_hostinfo.yaml')

        # Read files and checksums

        args_dict = vars(args)

        for item in ['files', 'checksum']:

            if args_dict[item]:

                key = list(filter(None, args_dict[item + '_key'].split('/')))
                filename = args_dict[item]

                if 'host' in data:
                    scp.from_remote(
                        data['host'],
                        os.path.join(data['prefix'], filename),
                        temp_file,
                        args.verbose)
                    filename = temp_file

                data[item] = yaml.read_item(filename, key)

        # Run basic checks

        if not args.ignore:
            for item in ['prefix']:
                if item not in data:
                    raise IOError('Please specify {0:s}'.format(item))
            if 'checksum' in data:
                if len(data['files']) != len(data['checksum']):
                    raise IOError('Number of checksums does not match number of files')

        # Clean-up: remove temporary file

        if 'host' in data:
            shutil.rmtree(temp_dir)

        # Write output

        if 'files' not in data:
            data['files'] = []

        yaml.dump(args.output, data, args.force)

    except Exception as e:

        print(e)
        return 1

if __name__ == '__main__':

    main()



'''shelephant_send
    Copy files to a remote, using earlier collected information on which files to copy where.
    By default the back-end is ``rysnc -a --from-file="temp" source_dir host:dest_dir``.
    Alternatively ``scp -p source_file host:dest_file`` can be used.
    Typically, ``rsync`` will be faster, especially in copying a lot of small files.

:usage:

    shelephant_send [options]

    shelephant_send [options] <files.yaml> <hostinfo.yaml>

:arguments:

    <files.yaml>
        YAML-file with files to send. Default: shelephant_dump.yaml

    <hostinfo.yaml>
        YAML-file with host information. Default: shelephant_hostinfo.yaml

:options:

    -k, --key=N
        Path in <files.yaml>, separated by "/". [default: /]

    -l, --local=N
        Add local 'host' information to use precomputed checksums.

    --scp
        Use ``scp`` instead of ``rysnc`` as backend.

    --check-rsync
        Check if files are different using *rsync*.
        *rsync* uses basic criteria such as file size and creation and modification date.
        This is much faster than using checksums but is only approximate.
        *rsync* can also check based on checksum.

    --temp=N
        Temporary filename to communicate with rsync. [default: shelephant_files.txt]

    --colors=M
        Select color scheme from: none, dark. [default: dark]

    -s, --summary
        Print summary (and no details unless specified).

    -d, --details
        Print details (and no summary unless specified).

    -f, --force
        Force overwrite of all existing (but not matching) files.

    --verbose
        Verbose all commands.

    -q, --quiet
        Do not print progress.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import os
import shutil

from .. import detail
from .. import rsync
from .. import scp
from .. import version
from .. import yaml


def main():

    try:

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-k', '--key', required=False, default='/')
        parser.add_argument('-l', '--local', required=False)
        parser.add_argument(      '--scp', required=False, action='store_true')
        parser.add_argument('-r', '--check-rsync', required=False, action='store_true')
        parser.add_argument(      '--temp', required=False, default='shelephant_files.txt')
        parser.add_argument(      '--colors', required=False, default='dark')
        parser.add_argument('-s', '--summary', required=False, action='store_true')
        parser.add_argument('-d', '--details', required=False, action='store_true')
        parser.add_argument('-f', '--force', required=False, action='store_true')
        parser.add_argument(      '--verbose', required=False, action='store_true')
        parser.add_argument('-q', '--quiet', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('args', nargs=2, default=['shelephant_dump.yaml', 'shelephant_hostinfo.yaml'])
        args = parser.parse_args()

        source = args.args[0]
        hostinfo = args.args[1]
        data = yaml.read(hostinfo)
        key = list(filter(None, args.key.split('/')))
        files = yaml.read_item(source, key)
        src_dir = os.path.dirname(source)
        dest_dir = data['prefix']

        if 'host' not in data:

            detail.copy(
                copy_function = shutil.copy2,
                files = files,
                src_dir = src_dir,
                dest_dir = data['prefix'],
                checksum = 'checksum' in data,
                check_rsync = None if not args.check_rsync else args.temp,
                quiet = args.quiet,
                force = args.force,
                print_details = not (args.force or args.summary) or args.details,
                print_summary = not (args.force or args.details) or args.summary,
                print_all = args.details,
                theme_name = args.colors.lower(),
                yaml_hostinfo_src = args.local,
                yaml_hostinfo_dest = hostinfo)

        elif args.scp:

            detail.copy_ssh(
                copy_function = scp.to_remote,
                use_rsync = False,
                host = data['host'],
                files = files,
                src_dir = src_dir,
                dest_dir = dest_dir,
                to_remote = True,
                checksum = 'checksum' in data,
                check_rsync = None if not args.check_rsync else args.temp,
                quiet = args.quiet,
                force = args.force,
                print_details = not (args.force or args.summary) or args.details,
                print_summary = not (args.force or args.details) or args.summary,
                print_all = args.details,
                verbose = args.verbose,
                theme_name = args.colors.lower(),
                yaml_hostinfo_src = args.local,
                yaml_hostinfo_dest = hostinfo)

        else:

            detail.copy_ssh(
                copy_function = rsync.to_remote,
                use_rsync = True,
                host = data['host'],
                files = files,
                src_dir = src_dir,
                dest_dir = dest_dir,
                to_remote = True,
                checksum = 'checksum' in data,
                check_rsync = None if not args.check_rsync else args.temp,
                quiet = args.quiet,
                force = args.force,
                print_details = not (args.force or args.summary) or args.details,
                print_summary = not (args.force or args.details) or args.summary,
                print_all = args.details,
                verbose = args.verbose,
                theme_name = args.colors.lower(),
                yaml_hostinfo_src = args.local,
                yaml_hostinfo_dest = hostinfo,
                tempfilename = args.temp)

    except Exception as e:

        print(e)
        return 1


if __name__ == '__main__':

    main()



'''ssh_test
    Run SSH test.

Usage:
    ssh_test [options] --host=N --prefix=N

Options:
    -r, --host=N        Host name.
    -p, --prefix=M      Path on host.
        --version       Print version.
    -g, --help          Print this help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import subprocess
import docopt
import os
import shelephant


def run(cmd):
    print(cmd)
    return subprocess.check_output(cmd, shell=True).decode('utf-8')


args = docopt.docopt(__doc__, version=shelephant.version)

# shelephant_send - basic

operations = [
    'bar.txt -> bar.txt',
    'foo.txt == foo.txt',
]

output = run(('shelephant_hostinfo -o myssh_send/shelephant_hostinfo.yaml --force '
              '--host "{0:s}" --prefix "{1:s}" -f -c').format(
               args['--host'], os.path.join(args['--prefix'], 'myssh_get')))

output = run(('shelephant_send --detail --colors none --force '
              'myssh_send/shelephant_dump.yaml myssh_send/shelephant_hostinfo.yaml'))

output = list(filter(None, output.split('\n')))
assert output == operations

# shelephant_send - local checksum

operations = [
    'bar.txt -> bar.txt',
    'foo.txt == foo.txt',
]

output = run(('shelephant_hostinfo -o myssh_send/shelephant_hostinfo.yaml --force '
              '--host "{0:s}" --prefix "{1:s}" -f -c').format(
              args['--host'], os.path.join(args['--prefix'], 'myssh_get')))

output = run(('shelephant_hostinfo -o myssh_send/shelephant_local.yaml --force '
              '-f myssh_send/shelephant_dump.yaml -c myssh_send/shelephant_checksum.yaml'))

output = run(('shelephant_send --detail --colors none --force '
              'myssh_send/shelephant_dump.yaml myssh_send/shelephant_hostinfo.yaml'))

output = list(filter(None, output.split('\n')))
assert output == operations

# shelephant_get - basic

operations = [
    'bar.txt -> bar.txt',
    'foo.txt == foo.txt',
]

output = run(('shelephant_hostinfo -o myssh_get/shelephant_hostinfo.yaml --force '
              '--host "{0:s}" --prefix "{1:s}" -f -c').format(
              args['--host'], os.path.join(args['--prefix'], 'myssh_send')))

output = run('shelephant_get --detail --colors none --force myssh_get/shelephant_hostinfo.yaml')

output = list(filter(None, output.split('\n')))
assert output == operations

os.remove('myssh_get/bar.txt')




import yaml
import functools
import operator
import click
import os


def read(filename):
    r'''
Read YAML file and return its content as ``list`` or ``dict``.
    '''

    if not os.path.isfile(filename):
        raise IOError('"{0:s} does not exist'.format(filename))

    with open(filename, 'r') as file:
        return yaml.load(file.read(), Loader=yaml.FullLoader)


def read_item(filename, key=[]):
    r'''
Get an item from a YAML file.
Optionally the key to the item can be specified as a list. E.g.
*   ``[]`` for a YAML file containing only a list.
*   ``['foo']`` for a plain YAML file.
*   ``['key', 'to', foo']`` for a YAML file with nested items.
    '''

    data = read(filename)

    if len(key) == 0 and type(data) != list:
        raise IOError('Specify key for "{0:s}"'.format(filename))

    if len(key) > 0:
        try:
            return functools.reduce(operator.getitem, key, data)
        except:
            raise IOError('"{0:s}" not in "{1:s}"'.format('/'.join(key), filename))

    return data


def dump(filename, data, force=False):
    r'''
Dump data to YAML file.

:type filename: str
:param filename: The output filename.

:type data: list, dict
:param data: The data to dump.

:type force: bool, optional
:param force: Do not prompt to overwrite file.
    '''

    dirname = os.path.dirname(filename)

    if not force:
        if os.path.isfile(filename):
            if not click.confirm('Overwrite "{0:s}"?'.format(filename)):
                raise IOError('Cancelled')
        elif not os.path.isdir(dirname) and len(dirname) > 0:
            if not click.confirm('Create "{0:s}"?'.format(os.path.dirname(filename))):
                raise IOError('Cancelled')

    if not os.path.isdir(dirname) and len(dirname) > 0:
        os.makedirs(os.path.dirname(filename))

    with open(filename, 'w') as file:
        ret = yaml.dump(data, file)


def view(data):
    r'''
Print data formatted as YAML.
    '''
    print(yaml.dump(data, default_flow_style=False, default_style=''))





from .external import exec_cmd

def from_remote(host, source, dest, verbose=False):
    r'''
Copy a file from a remote system. Uses ``scp -p``.
    '''

    cmd = 'scp -p {host:s}:{source:s} {dest:s}'.format(
        host=host, source=source, dest=dest)

    exec_cmd(cmd, verbose)


def to_remote(host, source, dest, verbose=False):
    r'''
Copy a file to a remote system. Uses ``scp -p``.
    '''

    cmd = 'scp -p {source:s} {host:s}:{dest:s}'.format(
        host=host, source=source, dest=dest)

    exec_cmd(cmd, verbose)





'''shelephant_mv
    Move files listed in a (field of a) YAML-file.
    The filenames are assumed either absolute, or relative to the input YAML-file.

:usage:

    shelephant_mv [options] <destination>

    shelephant_mv [options] <input.yaml> <destination>

:argument:

    <input.yaml>
        YAML-file with filenames. Default: shelephant_dump.yaml

    <destination>
        Prefix of the destination.

:options:

    -c, --checksum
        Use checksum to skip files that are the same.

    -k, --key=N
        Path in the YAML-file, separated by "/". [default: /]

    --colors=M
        Select color scheme from: none, dark. [default: dark]

    -s, --summary
        Print summary (and no details unless specified).

    -d, --details
        Print details (and no summary unless specified).

    -f, --force
        Move without prompt.

    -q, --quiet
        Do not print progress.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import os

from .. import detail
from .. import version
from .. import yaml


def main():

    try:

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-c', '--checksum', required=False, action='store_true')
        parser.add_argument('-k', '--key', required=False, default='/')
        parser.add_argument(      '--colors', required=False, default='dark')
        parser.add_argument('-s', '--summary', required=False, action='store_true')
        parser.add_argument('-d', '--details', required=False, action='store_true')
        parser.add_argument('-f', '--force', required=False, action='store_true')
        parser.add_argument('-q', '--quiet', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('args', nargs='+')
        args = parser.parse_args()

        if len(args.args) == 1:
            source = 'shelephant_dump.yaml'
            dest_dir = args.args[0]
        elif len(args.args) == 2:
            source = args.args[0]
            dest_dir = args.args[1]
        else:
            raise IOError('Too many arguments specified')

        key = list(filter(None, args.key.split('/')))

        return detail.copy(
            copy_function = os.rename,
            files = yaml.read_item(source, key),
            src_dir = os.path.dirname(source),
            dest_dir = dest_dir,
            checksum = args.checksum,
            quiet = args.quiet,
            force = args.force,
            print_details = not (args.force or args.summary) or args.details,
            print_summary = not (args.force or args.details) or args.summary,
            print_all = args.details,
            theme_name = args.colors.lower())

    except Exception as e:

        print(e)
        return 1


if __name__ == '__main__':

    main()




from collections import defaultdict
import click
import os

def _to_tree(d):
    r'''
Detail for: :py:fun:`filter_deepest`.
Not part of API.
    '''

    t = defaultdict(list)
    for a, *b in d:
        t[a].append(b)
    return {a: None if not (k := list(filter(None, b))) else _to_tree(k) for a, b in t.items()}

def _get_deepest_paths(d, c=[]):
    r'''
Detail for: :py:fun:`filter_deepest`.
Not part of API.

See https://stackoverflow.com/a/66211932/2646505
    '''
    for a, b in d.items():
        if b is None:
            yield '/'.join(c + [a])
        else:
            yield from _get_deepest_paths(b, c + [a])


def filter_deepest(paths):
    r'''
Return list with only the deepest paths.

For example::

    filter_deepest(["foo/bar/dir", "foo/bar"])
    >>> ["foo/bar/dir"]

:type paths: list of str
:param paths: List of paths.

:rtype: list of str
:return: List of paths.
    '''

    return list(_get_deepest_paths(_to_tree([i.split('/') for i in paths])))


def check_allisfile(paths):
    r'''
Check that all paths point to existing files.
Uses: ``os.path.isfile``.

:param list paths: List of file paths.
:throw: IOError
    '''

    if type(paths) == str:
        paths = [paths]

    for path in paths:
        if not os.path.isfile(path):
            raise IOError('"{0:s}" does not exist'.format(path))


def dirnames(paths, return_unique=True):
    r'''
Get the ``os.path.dirname`` of all file paths.

:param list paths: List of file paths.
:param bool return_unique: Filter duplicates.
:return: List of dirnames.
    '''

    if type(paths) == str:
        paths = [paths]

    ret = [os.path.dirname(path) for path in paths]

    if not return_unique:
        return ret

    return list(set(ret))


def overwrite(paths, force=False):
    r'''
List files that will be overwritten.

:param list paths: List of file paths.
:param bool force: If ``True`` the user is prompted to overwrite.
:return: List of overwritten files.
    '''

    if type(paths) == str:
        paths = [paths]

    ret = [path for path in paths if os.path.isfile(path)]

    if force or len(ret) == 0:
        return ret

    print('Files exist:')
    print('\n'.join(ret))
    if not click.confirm('Overwrite?'):
        raise IOError('Cancelled')

    return ret


def makedirs(dirnames, force=False):
    r'''
Same as :py:mod:`shelephant.MakeDir` but for list of directories.

:param bool force: Create directories without prompt.
    '''

    dirnames = [dirname for dirname in dirnames if len(dirname) > 0]
    dirnames = [dirname for dirname in dirnames if not os.path.isdir(dirname)]

    if len(dirnames) == 0:
        return 0

    dirnames = list(set(dirnames))
    dirnames = sorted(GetDeepestPaths(dirnames))

    if not force:
        for dirname in dirnames:
            print('mkdir -p {0:s}'.format(dirname))
        if not click.confirm('Proceed?'):
            raise IOError('Cancelled')

    for dirname in dirnames:
        os.makedirs(dirname)






import subprocess

def exec_cmd(cmd, verbose=False):
    r'''
Run command, optionally verbose command and its output, and return output.

:type cmd: str
:param cmd: The command to run.

:type verbose: bool
:param verbose: Print command and its output.
    '''

    if verbose:
        print(cmd)

    ret = subprocess.check_output(cmd, shell=True).decode('utf-8')

    if verbose:
        print(ret)

    return ret



import os
import shutil
import subprocess


def run(cmd):
    return subprocess.check_output(cmd, shell=True).decode('utf-8')


for dirname in ['myssh_send', 'myssh_get']:

    if os.path.isdir(dirname):
        shutil.rmtree(dirname)

    os.mkdir(dirname)

with open('myssh_send/bar.txt', 'w') as file:
    file.write('bar')

with open('myssh_send/foo.txt', 'w') as file:
    file.write('foo')

with open('myssh_get/foo.txt', 'w') as file:
    file.write('foo')

output = run('shelephant_dump -o myssh_send/shelephant_dump.yaml myssh_send/bar.txt myssh_send/foo.txt')
output = run('shelephant_dump -o myssh_get/shelephant_dump.yaml myssh_get/foo.txt')
output = run('shelephant_checksum -o myssh_send/shelephant_checksum.yaml myssh_send/shelephant_dump.yaml')
output = run('shelephant_checksum -o myssh_get/shelephant_checksum.yaml myssh_get/shelephant_dump.yaml')





'''shelephant_merge
    Merge a YAML file into another YAML file.

    Unless you use --no-path, the function assumes that all data are paths,
    and changes all relative paths from being relative to <branch.yaml> or <main.yaml>
    to being relative to --output.

:usage:

    shelephant_merge [options] <branch.yaml> <main.yaml>

:arguments:

    <branch.yaml>
        File to merge into main.yaml.

    <main.yaml>
        Main source.

:options:

    -o, --output=N
        Output file. (default: <main.yaml>)

    --replace
        Replace fields in <main.yaml> that are also in <branch.yaml>. (default: append)

    --skip
        Skip fields in <branch.yaml> that are also in <main.yaml>.

    --no-path
        Do not interpret data as paths.

    -f, --force
        Overwrite output file without prompt.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import mergedeep
import os

from .. import relpath
from .. import version
from .. import yaml

def recursive_items(dictionary):
    for key, value in dictionary.items():
        if type(value) is dict:
            yield from recursive_items(value)
        else:
            yield (key, value)


def main():

    try:

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-o', '--output', required=False)
        parser.add_argument(      '--replace', required=False, action='store_true')
        parser.add_argument(      '--skip', required=False, action='store_true')
        parser.add_argument(      '--no-path', required=False, action='store_true')
        parser.add_argument('-f', '--force', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('branch')
        parser.add_argument('main')
        args = parser.parse_args()

        main = yaml.read(args.main)
        branch = yaml.read(args.branch)
        output = args.output if args.output else args.main
        output_dir = os.path.dirname(output)

        if not args.no_path:

            paths = [os.path.dirname(args.main), os.path.dirname(args.branch)]

            for var, path in zip([main, branch], paths):
                if type(var) == list:
                    relpath.chroot(var, path, output_dir, in_place=True)
                elif type(var) == dict:
                    for key, value in recursive_items(var):
                        relpath.chroot(value, path, output_dir, in_place=True)
                else:
                    raise IOError('Files have an incompatible structure')

        if type(main) == list and type(branch) == list:

            if args.skip:
                pass
            elif args.replace:
                main = branch
            else:
                main += branch

        elif type(main) == dict and type(branch) == dict:

            from mergedeep import merge, Strategy

            if args.skip:
                mergedeep.merge(branch, main, strategy=mergedeep.Strategy.REPLACE)
                main = branch
            elif args.replace:
                mergedeep.merge(main, branch, strategy=mergedeep.Strategy.REPLACE)
            else:
                mergedeep.merge(main, branch, strategy=mergedeep.Strategy.ADDITIVE)

        else:

            raise IOError('Files have an incompatible structure')

        yaml.dump(output, main, args.force)

    except Exception as e:

        print(e)
        return 1


if __name__ == '__main__':

    main()



import unittest
import subprocess
import os
import shutil
import numpy as np
import shelephant


def run(cmd, verbose=False):
    return subprocess.check_output(cmd, shell=True).decode('utf-8')


class Test_tools(unittest.TestCase):

    def test_flatten(self):

        arg = [1, [2, 2, 2], 4]
        ret = [1, 2, 2, 2, 4]

        self.assertEqual(ret, shelephant.convert.flatten(arg))

    def test_squash(self):

        arg = {'foo': [1, 2], 'bar': {'foo': [3, 4], 'bar': 5}}
        ret = [1, 2, 3, 4, 5]

        self.assertEqual(ret, shelephant.convert.squash(arg))

    def test_deepest_dirs(self):

        dirnames = [
            "foo/bar",
            "foo/bar/mydir",
            "bar/mydir",
            "bar" ,
            "bar/mydir2",
            "foo/bar/mydir/deep",
            "foo/bar/mydir/deep/deeper",
            "foo/bar/mydir/deep/also",
            "foo/shallow"]

        ret = shelephant.path.filter_deepest(dirnames)

        d = [
            "bar/mydir",
            "bar/mydir2",
            "foo/bar/mydir/deep/deeper",
            "foo/bar/mydir/deep/also",
            "foo/shallow"]

        self.assertEqual(sorted(ret), sorted(d))


class Test_checksum(unittest.TestCase):

    def test_basic(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_checksum -f -q shelephant_dump.yaml')
        data = shelephant.yaml.read_item('shelephant_checksum.yaml')

        keys = [
            '2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae',
            'fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9',
        ]

        self.assertEqual(data, keys)

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('shelephant_dump.yaml')
        os.remove('shelephant_checksum.yaml')

    def test_hybrid(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f foo.txt')
        output = run('shelephant_checksum -f -q')
        output = run('shelephant_hostinfo --force -f -c')

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_checksum -f -q -l shelephant_hostinfo.yaml')
        data = shelephant.yaml.read_item('shelephant_checksum.yaml')

        keys = [
            '2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae',
            'fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9',
        ]

        self.assertEqual(data, keys)

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('shelephant_dump.yaml')
        os.remove('shelephant_checksum.yaml')
        os.remove('shelephant_hostinfo.yaml')

    def test_recursive(self):

        letters = ['a', 'b', 'c', 'd', 'e', 'f', 'g']

        for letter in letters:
            with open('{0:s}.txt'.format(letter), 'w') as file:
                file.write(letter)

        files = ['{0:s}.txt'.format(letter) for letter in letters]

        keys = [
            'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb', # a
            '3e23e8160039594a33894f6564e1b1348bbd7a0088d42c4acb73eeaed59c009d', # b
            '2e7d2c03a9507ae265ecf5b5356885a53393a2029d241394997265a1a25aefc6', # c
            '18ac3e7343f016890c510e93f935261169d9e3f565436429830faf0934f4f8e4', # d
            '3f79bb7b435b05321651daefd374cdc681dc06faa65e374e38337b88ca046dea', # e
            '252f10c83610ebca1a059c0bae8255eba2f95be4d1d7bcfa89d7248a82d9f111', # f
            'cd0aa9856147b6c5b4ff2b7dfee5da20aa38253099ef1b4a64aced233c9afe29', # g
        ]

        output = run('shelephant_dump -f -s {0:s}'.format(' '.join(files)))
        output = run('shelephant_checksum -f -q')

        data = shelephant.yaml.read_item('shelephant_checksum.yaml')

        self.assertEqual(data, keys)

        ifiles = np.arange(len(files))
        np.random.shuffle(ifiles)

        output = run('shelephant_dump -f {0:s}'.format(' '.join([files[i] for i in ifiles[:3]])))
        output = run('shelephant_checksum -f -q')
        output = run('shelephant_hostinfo --force -f -c')

        np.random.shuffle(ifiles)

        output = run('shelephant_dump -f {0:s}'.format(' '.join([files[i] for i in ifiles])))
        output = run('shelephant_checksum -f -q -l shelephant_hostinfo.yaml')
        output = run('shelephant_hostinfo --force -f -c')

        output = run('shelephant_dump -f -s {0:s}'.format(' '.join(files)))
        output = run('shelephant_checksum -f -q -l shelephant_hostinfo.yaml')

        data = shelephant.yaml.read_item('shelephant_checksum.yaml')

        self.assertEqual(data, keys)

        for file in files:
            os.remove(file)

        os.remove('shelephant_dump.yaml')
        os.remove('shelephant_checksum.yaml')
        os.remove('shelephant_hostinfo.yaml')


class Test_dump(unittest.TestCase):

    def test_basic(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        for dirname in ['mydir']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mydir')

        with open('mydir/foo.txt', 'w') as file:
            file.write('foo')

        with open('mydir/bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f -s -o dump_1.yaml foo.txt bar.txt')
        output = run('shelephant_dump -f -s -o dump_2.yaml *.txt')
        output = run('shelephant_dump -f -s -o mydir/dump_3.yaml mydir/*.txt')

        with open('dump_1.yaml', 'r') as file:
            dump_1 = file.read()

        with open('dump_2.yaml', 'r') as file:
            dump_2 = file.read()

        with open('mydir/dump_3.yaml', 'r') as file:
            dump_3 = file.read()

        self.assertEqual(dump_1, dump_2)
        self.assertEqual(dump_1, dump_3)

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('dump_1.yaml')
        os.remove('dump_2.yaml')
        shutil.rmtree('mydir')

    def test_append(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        with open('foo.pdf', 'w') as file:
            file.write('foo')

        with open('bar.pdf', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_dump -a foo.pdf bar.pdf')

        self.assertEqual(shelephant.yaml.read('shelephant_dump.yaml'), ['foo.txt', 'bar.txt', 'foo.pdf', 'bar.pdf'])

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('foo.pdf')
        os.remove('bar.pdf')
        os.remove('shelephant_dump.yaml')


class Test_extract(unittest.TestCase):

    def test_single_path(self):

        data = {
            'foo' : ['foo.txt', 'bar.txt'],
            'bar' : ['foo.pdf', 'bar.pdf'],
            'key' : ['foo.key', 'bar.key'],
        }

        shelephant.yaml.dump('dump.yaml', data, force=True)

        output = run('shelephant_extract -f dump.yaml "foo"')

        self.assertEqual(shelephant.yaml.read('dump.yaml'), ['foo.txt', 'bar.txt'])

        os.remove('dump.yaml')

    def test_multiple_paths(self):

        data = {
            'foo' : ['foo.txt', 'bar.txt'],
            'bar' : ['foo.pdf', 'bar.pdf'],
            'key' : ['foo.key', 'bar.key'],
            'sub' : {
                'foo' : ['foo.txt', 'bar.txt'],
                'bar' : ['foo.pdf', 'bar.pdf'],
                'key' : ['foo.key', 'bar.key'],
            },
        }

        shelephant.yaml.dump('dump.yaml', data, force=True)

        output = run('shelephant_extract -f dump.yaml "/sub/foo" "foo"')

        self.assertEqual(shelephant.yaml.read('dump.yaml'), {'foo': ['foo.txt', 'bar.txt'], 'sub': {'foo': ['foo.txt', 'bar.txt']}})

        os.remove('dump.yaml')

    def test_multiple_paths_squash(self):

        data = {
            'foo' : ['foo.txt', 'bar.txt'],
            'bar' : ['foo.pdf', 'bar.pdf'],
            'key' : ['foo.key', 'bar.key'],
            'sub' : {
                'foo' : ['foo2.txt', 'bar2.txt'],
                'bar' : ['foo2.pdf', 'bar2.pdf'],
                'key' : ['foo2.key', 'bar2.key'],
            },
        }

        shelephant.yaml.dump('dump.yaml', data, force=True)

        output = run('shelephant_extract -f --squash dump.yaml "/sub/foo" "foo"')

        self.assertEqual(shelephant.yaml.read('dump.yaml'), ['foo2.txt', 'bar2.txt', 'foo.txt', 'bar.txt'])

        os.remove('dump.yaml')


class Test_merge(unittest.TestCase):

    def test_basic(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -o main.yaml foo.txt')
        output = run('shelephant_dump -o branch.yaml bar.txt')
        output = run('shelephant_merge -f branch.yaml main.yaml')

        self.assertEqual(shelephant.yaml.read('main.yaml'), ['foo.txt', 'bar.txt'])

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('main.yaml')
        os.remove('branch.yaml')

        os.mkdir('dira')
        os.mkdir('dirb')

        with open('dira/foo.txt', 'w') as file:
            file.write('foo')

        with open('dira/bar.txt', 'w') as file:
            file.write('bar')

        with open('dirb/foo.txt', 'w') as file:
            file.write('foo')

        with open('dirb/bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -o dira/dump.yaml dira/foo.txt dira/bar.txt')
        output = run('shelephant_dump -o dirb/dump.yaml dirb/foo.txt dirb/bar.txt')
        output = run('shelephant_merge -f dira/dump.yaml dirb/dump.yaml')

        self.assertEqual(shelephant.yaml.read('dirb/dump.yaml'), ['foo.txt', 'bar.txt', '../dira/foo.txt', '../dira/bar.txt'])

        shutil.rmtree('dira')
        shutil.rmtree('dirb')


class Test_hostinfo(unittest.TestCase):

    def test_basic(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f -s -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml')
        output = run('shelephant_get -f -q mydest/hostinfo.yaml')
        output = run('shelephant_dump -f -s -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')


    def test_remove(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        keys = [
            '2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae',
            'fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9',
        ]

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_checksum -f -q')
        output = run('shelephant_hostinfo --force -f -c')
        output = run('shelephant_hostinfo --force --remove bar.txt')
        data = shelephant.yaml.read('shelephant_hostinfo.yaml')

        self.assertEqual(data['files'], ['foo.txt'])
        self.assertEqual(data['checksum'], [keys[0]])

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('shelephant_dump.yaml')
        os.remove('shelephant_checksum.yaml')
        os.remove('shelephant_hostinfo.yaml')


class Test_get(unittest.TestCase):

    def test_basic(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt -> foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml')
        output = run('shelephant_get -f -d -q --colors none mydest/hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')

    def test_partial(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        shutil.copy('mysrc/foo.txt', 'mydest/foo.txt')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt == foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml')
        output = run('shelephant_get -f -d -q --colors none mydest/hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')

    def test_partial_localchecksum(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        with open('mysrc/car.txt', 'w') as file:
            file.write('car')

        with open('mysrc/dog.txt', 'w') as file:
            file.write('dog')

        shutil.copy('mysrc/foo.txt', 'mydest/foo.txt')
        shutil.copy('mysrc/dog.txt', 'mydest/dog.txt')

        operations = [
            'bar.txt -> bar.txt',
            'car.txt -> car.txt',
            'dog.txt == dog.txt',
            'foo.txt == foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml')
        output = run('shelephant_hostinfo -o mydest/local.yaml -f mydest/files.yaml -c mydest/checksum.yaml')
        output = run('shelephant_get -f -d -q --colors none -l mydest/local.yaml mydest/hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump -f -s -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')

    def test_partial_rsync(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        shutil.copy2('mysrc/foo.txt', 'mydest/foo.txt')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt == foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml')
        output = run('shelephant_get -f -d -q --colors none --check-rsync mydest/hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')
        os.remove('shelephant_files.txt')


class Test_send(unittest.TestCase):

    def test_basic(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        with open('mydest/foobar.txt', 'w') as file:
            file.write('foobar')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt -> foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_hostinfo --force -o hostinfo.yaml -f mydest/files.yaml -c mydest/checksum.yaml')
        output = run('shelephant_send -f -d -q --colors none mysrc/files.yaml hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        os.remove('mydest/foobar.txt')

        output = run('shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')
        os.remove('hostinfo.yaml')

    def test_empty_remote(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt -> foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_hostinfo --force -o hostinfo.yaml -p mydest')
        output = run('shelephant_send -f -d -q --colors none mysrc/files.yaml hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_checksum -f -q -o mysrc/checksum.yaml mysrc/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')
        os.remove('hostinfo.yaml')

    def test_partial(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        shutil.copy('mysrc/foo.txt', 'mydest/foo.txt')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt == foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_hostinfo --force -o hostinfo.yaml -f mydest/files.yaml -c mydest/checksum.yaml')
        output = run('shelephant_send -f -d -q --colors none mysrc/files.yaml hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')
        os.remove('hostinfo.yaml')

    def test_partial_localchecksum(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        with open('mysrc/car.txt', 'w') as file:
            file.write('car')

        with open('mysrc/dog.txt', 'w') as file:
            file.write('dog')

        shutil.copy('mysrc/foo.txt', 'mydest/foo.txt')
        shutil.copy('mysrc/dog.txt', 'mydest/dog.txt')

        operations = [
            'bar.txt -> bar.txt',
            'car.txt -> car.txt',
            'dog.txt == dog.txt',
            'foo.txt == foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_hostinfo --force -o hostinfo.yaml -f mydest/files.yaml -c mydest/checksum.yaml')
        output = run('shelephant_hostinfo --force -o local.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml')
        output = run('shelephant_send -f -d -q --colors none -l local.yaml mysrc/files.yaml hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')
        os.remove('hostinfo.yaml')
        os.remove('local.yaml')

    def test_partial_rsync(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        shutil.copy2('mysrc/foo.txt', 'mydest/foo.txt')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt == foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_hostinfo --force -o hostinfo.yaml -p mydest')
        output = run('shelephant_send -f -d -q --colors none --check-rsync mysrc/files.yaml hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')
        os.remove('hostinfo.yaml')
        os.remove('shelephant_files.txt')


class Test_mv(unittest.TestCase):

    def test_basic(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_mv -f -q mysrc/files.yaml mydest')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')


class Test_cp(unittest.TestCase):

    def test_basic(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_cp -f -q mysrc/files.yaml mydest')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))
        self.assertTrue(os.path.isfile('mysrc/foo.txt'))
        self.assertTrue(os.path.isfile('mysrc/bar.txt'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')

    def test_rsync(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.log', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.log', 'w') as file:
            file.write('bar')

        shutil.copy2('mysrc/foo.log', 'mydest/foo.log')

        operations = [
            'bar.log -> bar.log',
            'foo.log == foo.log',
        ]

        output = run('shelephant_dump -o mysrc/files.yaml mysrc/*.log')
        output = run('shelephant_cp -f -d -q --colors none --check-rsync mysrc/files.yaml mydest')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.log')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')

        self.assertEqual(shelephant.yaml.read('mysrc/files.yaml'), shelephant.yaml.read('mydest/files.yaml'))
        self.assertEqual(shelephant.yaml.read('mysrc/checksum.yaml'), shelephant.yaml.read('mydest/checksum.yaml'))
        self.assertTrue(os.path.isfile('mysrc/foo.log'))
        self.assertTrue(os.path.isfile('mysrc/bar.log'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')
        os.remove('shelephant_files.txt')


class Test_rm(unittest.TestCase):

    def test_basic(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_rm -f shelephant_dump.yaml')

        self.assertFalse(os.path.isfile('foo.txt'))
        self.assertFalse(os.path.isfile('bar.txt'))

        os.remove('shelephant_dump.yaml')


class Test_parse(unittest.TestCase):

    def test_basic(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_parse shelephant_dump.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), ['- foo.txt', '- bar.txt'])

        os.remove('shelephant_dump.yaml')
        os.remove('foo.txt')
        os.remove('bar.txt')


if __name__ == '__main__':

    unittest.main()



r'''
Implementation details.
Not part of public API.

(c) Tom de Geus, 2021, MIT
'''


import click
import numpy as np
import os
import tqdm

from .checksum import get
from .path import makedirs
from .relpath import add_prefix
from .rich import String
from .rich import theme
from .rsync import diff
from .yaml import read


def copy(
    copy_function,
    files,
    src_dir,
    dest_dir,
    checksum = False,
    check_rsync = None,
    quiet = False,
    force = False,
    print_details = True,
    print_summary = True,
    print_all = False,
    theme_name = 'none',
    yaml_hostinfo_src = None,
    yaml_hostinfo_dest = None):
    r'''
Copy/move files.

:param function copy_function:
    Function to perform the copy.
    Called with ``copy_function(source_path, dest_path)``.
    E.g. ``os.rename``.

:type files: list of str
:param files: Filenames (will be prepended by ``src_dir`` and ``dest_dir``).

:param str src_dir: The source directory.
:param str dest_dir: The destination directory.
:param bool checksum: Use checksum to skip files that are the same.
:param str check_rsync: Use rsync to check files to skip, specify temporary file name.
:param bool quiet: Proceed without printing progress.
:param bool force: Continue without prompt.
:param bool print_details: Print copy details.
:param bool print_summary: Print copy summary.
:param bool print_all: If ``False`` auto-truncation of long output is applied.

:type theme_name: str or None
:param theme_name: The name of the color-theme. See :py:mod:`shelephant.theme`.

:type yaml_hostinfo_src: str, optional
:param yaml_hostinfo_src:
    Filename of hostinfo for the source, see :py:mod:`shelephant.cli.hostinfo`.
    Specify these files *only* to use precomputed checksums.

:type yaml_hostinfo_dest: str, optional
:param yaml_hostinfo_dest:
    Filename of hostinfo for the destination, see :py:mod:`shelephant.cli.hostinfo`.
    Specify these files *only* to use precomputed checksums.
    '''

    assert type(files) == list

    if len(files) == 0:
        if not quiet:
            print('Nothing to do')
        return

    src = add_prefix(src_dir, files)
    dest = add_prefix(dest_dir, files)
    n = len(src)
    overwrite = [False for i in range(n)]
    create = [False for i in range(n)]
    skip = [False for i in range(n)]
    color = theme(theme_name.lower())

    for file in src:
        if not os.path.isfile(file):
            raise IOError('Input file "{0:s}" does not exists'.format(file))

    if not os.path.isdir(dest_dir) and len(dest_dir) > 0:

        create = [True for i in range(n)]

    elif check_rsync is not None:

        tmp = diff(
            source_dir = src_dir,
            dest_dir = dest_dir,
            tempfilename = check_rsync,
            files = files,
            checksum = checksum,
            force = force)

        skip = tmp['skip']
        create = tmp['create']
        overwrite = tmp['overwrite']

    else:

        if checksum:
            src_checksums = get(src, yaml_hostinfo_src, progress=not quiet)
            dest_checksums = get(dest, yaml_hostinfo_dest, progress=not quiet)

        for i in range(n):
            if os.path.isfile(dest[i]):
                if checksum:
                    if (src_checksums[i] == dest_checksums[i]) and (src_checksums[i] is not None):
                        skip[i] = True
                        continue
                overwrite[i] = True
                continue
            create[i] = True

    l = max([len(file) for file in files])
    ncreate = sum(create)
    noverwrite = sum(overwrite)
    nskip = sum(skip)
    pcreate = not ((noverwrite > 0 and ncreate > 100) or ncreate > 300) or print_all
    pskip = (nskip <= 100) or print_all
    pcreate_message = ' (not printed)' if not pcreate else ''
    pskip_message = ' (not printed)' if not pskip else ''

    overview = []
    if ncreate > 0:
        overview += [String('create (->): {0:d}{1:s}'.format(ncreate, pcreate_message), color=color['new']).format()]
    if noverwrite > 0:
        overview += [String('overwrite (=>): {0:d}'.format(noverwrite), color=color['overwrite']).format()]
    if nskip > 0:
        overview += [String('skip (==): {0:d}{1:s}'.format(nskip, pskip_message), color=color['skip']).format()]

    summary = []
    summary += ['- source : ' + os.path.normpath(src_dir)]
    summary += ['- dest   : ' + os.path.normpath(dest_dir)]
    summary += ['- ' + ', '.join(overview)]

    if ncreate + noverwrite <= 100 and print_summary:
        print('-----')
        print('\n'.join(summary))
        print('-----')

    if print_details:
        for i in range(n):
            if create[i] and pcreate:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=color['bright']).format(),
                    String('->', color=color['bright']).format(),
                    String(files[i], color=color['new']).format()
                ))
            elif skip[i] and pskip:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=color['skip']).format(),
                    String('==', color=color['skip']).format(),
                    String(files[i], color=color['skip']).format()
                ))
            elif overwrite[i]:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=color['bright']).format(),
                    String('=>', color=color['bright']).format(),
                    String(files[i], color=color['overwrite']).format()
                ))

    if ncreate + noverwrite > 100 and print_summary:
        print('-----')
        print('\n'.join(summary))
        print('-----')

    if all(skip):
        return 0

    if not force:
        if not click.confirm('Proceed?'):
            raise IOError('Cancelled')

    makedirs(list(set([os.path.dirname(i) for i in dest])))

    i = np.argwhere(np.not_equal(skip, True)).ravel()
    src = np.array(src)[i]
    dest = np.array(dest)[i]
    files = np.array(files)[i]

    for i in tqdm.trange(len(files), disable=quiet):
        copy_function(src[i], dest[i])


def copy_ssh(
    copy_function,
    use_rsync,
    host,
    files,
    src_dir,
    dest_dir,
    to_remote,
    checksum = False,
    check_rsync = None,
    quiet = False,
    force = False,
    print_details = True,
    print_summary = True,
    print_all = False,
    verbose = False,
    theme_name = 'none',
    yaml_hostinfo_src = None,
    yaml_hostinfo_dest = None,
    tempfilename = None):
    r'''
Get/send files.

:param function copy_function:
    Function to perform the copy.
    Called differently depending on ``use_rsync``.

:param bool use_rsync: Signal if ``copy_function`` uses *rsync*.
:param str host: Host-name

:type files: list of str
:param files: Filenames (will be prepended by ``src_dir`` and ``dest_dir``).

:param str src_dir: The source directory.
:param str dest_dir: The destination directory.
:param bool to_remote: Sets direction of copy operation.
:param bool checksum: Use checksum to skip files that are the same.
:param str check_rsync: Use rsync to check files to skip, specify temporary file name.
:param bool quiet: Proceed without printing progress.
:param bool force: Continue without prompt.
:param bool print_details: Print copy details.
:param bool print_summary: Print copy summary.
:param bool print_all: If ``False`` auto-truncation of long output is applied.

:type theme_name: str or None
:param theme_name: The name of the color-theme. See :py:fun:`shelephant.rich.theme`.

:type yaml_hostinfo_src: str, optional
:param yaml_hostinfo_src:
    Filename of hostinfo for the source, see :py:mod:`shelephant.cli.hostinfo`.
    Specify these files *only* to use precomputed checksums.

:type yaml_hostinfo_dest: str, optional
:param yaml_hostinfo_dest:
    Filename of hostinfo for the destination, see :py:mod:`shelephant.cli.hostinfo`.
    Specify these files *only* to use precomputed checksums.

:param str tempfilename: Filename for temporary file to use (e.g. for ``rsync``).
    '''

    assert type(files) == list

    if len(files) == 0:
        if not quiet:
            print('Nothing to do')
        return

    src = add_prefix(src_dir, files)
    dest = add_prefix(dest_dir, files)
    n = len(src)
    overwrite = [False for i in range(n)]
    create = [False for i in range(n)]
    skip = [False for i in range(n)]
    dest_exists = [False for i in range(n)]
    color = theme(theme_name.lower())

    if check_rsync is not None:

        tmp = diff(
            source_dir = src_dir if to_remote else hostname + ":" + src_dir,
            dest_dir = hostname + ":" + dest_dir if to_remote else dest_dir,
            tempfilename = check_rsync,
            files = files,
            checksum = checksum,
            force = force)

        skip = tmp['skip']
        create = tmp['create']
        overwrite = tmp['overwrite']

    else:

        if checksum == True:
            src_checksums = get(src, yaml_hostinfo_src, progress=not quiet)
            dest_checksums = get(dest, yaml_hostinfo_dest, progress=not quiet)

        if to_remote:
            if yaml_hostinfo_dest:
                f = read(yaml_hostinfo_dest)['files']
                for i in range(n):
                    if files[i] in f:
                        dest_exists[i] = True
        elif not to_remote:
            for i in range(n):
                if os.path.isfile(dest[i]):
                    dest_exists[i] = True

        for i in range(n):
            if dest_exists[i]:
                if checksum:
                    if (src_checksums[i] == dest_checksums[i]) and (src_checksums[i] is not None):
                        skip[i] = True
                        continue
                overwrite[i] = True
                continue
            create[i] = True

    l = max([len(file) for file in files])
    ncreate = sum(create)
    noverwrite = sum(overwrite)
    nskip = sum(skip)
    pcreate = not ((noverwrite > 0 and ncreate > 100) or ncreate > 300) or print_all
    pskip = (nskip <= 100) or print_all
    pcreate_message = ' (not printed)' if not pcreate else ''
    pskip_message = ' (not printed)' if not pskip else ''

    overview = []
    if ncreate > 0:
        overview += [String('create (->): {0:d}{1:s}'.format(ncreate, pcreate_message), color=color['new']).format()]
    if noverwrite > 0:
        overview += [String('overwrite (=>): {0:d}'.format(noverwrite), color=color['overwrite']).format()]
    if nskip > 0:
        overview += [String('skip (==): {0:d}{1:s}'.format(nskip, pskip_message), color=color['skip']).format()]

    summary = []
    if to_remote:
        summary += ['- to host        : ' + host]
        summary += ['- source (local) : ' + os.path.normpath(src_dir)]
        summary += ['- dest (remote)  : ' + os.path.normpath(dest_dir)]
    else:
        summary += ['- from host       : ' + host]
        summary += ['- source (remote) : ' + os.path.normpath(src_dir)]
        summary += ['- dest (local)    : ' + os.path.normpath(dest_dir)]
    summary += ['- ' + ', '.join(overview)]

    if ncreate + noverwrite <= 100 and print_summary:
        print('-----')
        print('\n'.join(summary))
        print('-----')

    if print_details:
        for i in range(n):
            if create[i] and pcreate:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=color['bright']).format(),
                    String('->', color=color['bright']).format(),
                    String(files[i], color=color['new']).format()
                ))
            elif skip[i] and pskip:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=color['skip']).format(),
                    String('==', color=color['skip']).format(),
                    String(files[i], color=color['skip']).format()
                ))
            elif overwrite[i]:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=color['bright']).format(),
                    String('=>', color=color['bright']).format(),
                    String(files[i], color=color['overwrite']).format()
                ))

    if ncreate + noverwrite > 100 and print_summary:
        print('-----')
        print('\n'.join(summary))
        print('-----')

    if all(skip):
        return 0

    if not force:
        if not click.confirm('Proceed?'):
            raise IOError('Cancelled')

    if not to_remote:
        makedirs(list(set([os.path.dirname(i) for i in dest])))

    i = np.argwhere(np.not_equal(skip, True)).ravel()
    src = np.array(src)[i]
    dest = np.array(dest)[i]
    files = np.array(files)[i]

    if use_rsync:
        dest_dir = dest_dir if len(dest_dir) > 0 else '.'
        src_dir = src_dir if len(src_dir) > 0 else '.'
        return copy_function(host, src_dir, dest_dir, tempfilename, files, force, verbose, not quiet)

    pbar = tqdm.trange(len(files), disable=quiet)
    for i in pbar:
        pbar.set_description(files[i])
        copy_function(host, src[i], dest[i], verbose)




import click
import numpy as np
import os
import re
import subprocess
import tqdm

from .external import exec_cmd

def _rsync(
    source_dir,
    dest_dir,
    tempfilename,
    files,
    force=False,
    verbose=False,
    progress=True):
    r'''
Copy files to a destination using ``rsync -a --files-from``.

:param str source_dir: Source directory.
:param str dest_dir: Source directory.
:param str tempfilename: Path of a temporary file to use to direct ``rsync --files-from``.
:param list files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
:param bool force: Continue without prompt.
:param bool verbose: Verbose commands.
:param bool progress: Show progress bar.
    '''

    assert type(tempfilename) == str

    if not force:
        if os.path.isfile(tempfilename):
            if not click.confirm('Overwrite "{0:s}"?'.format(tempfilename)):
                raise IOError('Cancelled')

    open(tempfilename, 'w').write('\n'.join(files))

    # Run without printing output

    if not progress:

        cmd = 'rsync -a --files-from="{files:s}" "{source_dir:s}" "{dest_dir:s}"'.format(
            source_dir=source_dir, dest_dir=dest_dir, files=tempfilename)

        return exec_cmd(cmd, verbose)

    # Run while printing output

    cmd = 'rsync -aP --files-from="{files:s}" "{source_dir:s}" "{dest_dir:s}"'.format(
        source_dir=source_dir, dest_dir=dest_dir, files=tempfilename)

    if verbose:
        print(cmd)

    pbar = tqdm.tqdm(total=len(files))
    sbar = tqdm.tqdm(unit='B', unit_scale=True)

    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)

    for line in iter(process.stdout.readline, b''):
        line = line.decode("utf-8")
        if re.match(r'(.*)(xf)([e]?)(r\#)([0-9])(.*)(to\-ch)([e]?[c]?)(k\=)([0-9])(.*)', line):
            e = int(list(filter(None, line.split(" ")))[-6].replace(",", ""))
            pbar.update()
            sbar.update(e)


def from_remote(
    hostname,
    source_dir,
    dest_dir,
    tempfilename,
    files,
    force=False,
    verbose=False,
    progress=True):
    r'''
Copy files to a remote system using ``rsync -a --files-from``.

:param str hostname: Hostname.
:param str source_dir: Source directory.
:param str dest_dir: Source directory.
:param str tempfilename: Path of a temporary file to use to direct ``rsync --files-from``.
:param list files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
:param bool force: Continue without prompt.
:param bool verbose: Verbose commands.
:param bool progress: Show progress bar.
    '''

    return _rsync(
        source_dir = hostname + ":" + source_dir,
        dest_dir = dest_dir,
        tempfilename = tempfilename,
        files = files,
        force = force,
        verbose = verbose,
        progress = progress)


def to_remote(
    hostname,
    source_dir,
    dest_dir,
    tempfilename,
    files,
    force=False,
    verbose=False,
    progress=True):
    r'''
Copy files from a remote system using ``rsync -a --files-from``.

:param str hostname: Hostname.
:param str source_dir: Source directory.
:param str dest_dir: Source directory.
:param str tempfilename: Path of a temporary file to use to direct ``rsync --files-from``.
:param list files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
:param bool force: Continue without prompt.
:param bool verbose: Verbose commands.
:param bool progress: Show progress bar.
    '''

    return _rsync(
        source_dir = source_dir,
        dest_dir = hostname + ":" + dest_dir,
        tempfilename = tempfilename,
        files = files,
        force = force,
        verbose = verbose,
        progress = progress)


def diff(
    source_dir,
    dest_dir,
    tempfilename,
    files,
    checksum=False,
    force=False,
    verbose=False):
    r'''
Check if files are different using *rsync*.
*rsync* uses basic criteria such as file size and creation and modification date.
This is much faster than using checksums but is only approximate.
See `rsync manual <https://www.samba.org/ftp/rsync/rsync.html>`_.

:param str source_dir: Source directory.
:param str dest_dir: Source directory.
:param str tempfilename: Path of a temporary file to use to direct ``rsync --files-from``.
:param list files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
:param bool checksum: Use checksum to test file difference.
:param bool force: Continue without prompt.
:param bool verbose: Verbose commands.
    '''

    files = [os.path.normpath(file) for file in files]

    assert type(tempfilename) == str

    if not force:
        if os.path.isfile(tempfilename):
            if not click.confirm('Overwrite "{0:s}"?'.format(tempfilename)):
                raise IOError('Cancelled')

    open(tempfilename, 'w').write('\n'.join(files))

    # Run without printing output

    opt = '-nai'

    if checksum:
        opt += 'c'

    cmd = 'rsync {opt:s} --files-from="{files:s}" "{source_dir:s}" "{dest_dir:s}"'.format(
        source_dir=source_dir, dest_dir=dest_dir, files=tempfilename, opt=opt)

    lines = list(filter(None, exec_cmd(cmd, verbose).split('\n')))
    lines = [line for line in lines if line[1] == 'f']

    if len(lines) == 0:
        return {
            'skip' : np.ones((len(files)), dtype=np.bool),
            'create' : np.zeros((len(files)), dtype=np.bool),
            'overwrite' : np.zeros((len(files)), dtype=np.bool),
        }

    check_paths = [line.split(' ')[1] for line in lines]
    mode = np.zeros((len(check_paths)), dtype=np.int16)

    for i, line in enumerate(lines):
        if line[0] == '>':
            if line[2] == '+':
                mode[i] = 1 # create
            else:
                mode[i] = 2 # overwrite
        elif line[0] == '.':
            pass
        else:
            raise IOError('Unknown cryptic output "{0:s}"'.format(line))

    sorter = np.argsort(files)
    source_paths = np.array(files, dtype=str)[sorter]

    i = np.argsort(check_paths)
    check_paths = np.array(check_paths, dtype=str)[i]
    mode = mode[i]

    test = np.in1d(source_paths, check_paths)

    idx = np.searchsorted(check_paths, source_paths)
    idx = np.where(test, idx, 0)
    ret = np.where(test, mode[idx], 0)
    ret = ret.astype(np.int16)
    out = np.empty_like(ret)
    out[sorter] = ret

    return {
        'skip' : out == 0,
        'create' : out == 1,
        'overwrite' : out == 2,
    }




from .external import exec_cmd


def file_exists(host, source, verbose=False):
    r'''
Check if a file exists on a remote system. Uses ``ssh``.
    '''

    cmd = 'ssh {host:s} test -f "{source:s}" && echo found || echo not found'.format(
        host=host, source=source)

    ret = exec_cmd(cmd, args['--verbose'])

    if ret == 'found':
        return True

    return False





def theme(theme=None):
    r'''
Return dictionary of colors.

.. code-block:: python

    {
        'new' : '...',
        'overwrite' : '...',
        'skip' : '...',
        'bright' : '...',
    }

:param str theme: Select color-theme.

:rtype: dict
    '''

    if theme == 'dark':
        return \
        {
            'new' : '1;32',
            'overwrite': '1;31',
            'skip' : '1;30',
            'bright' : '1;37',
        }

    return \
    {
        'new' : '',
        'overwrite': '',
        'skip' : '',
        'bright' : '',
    }


class String:
    r'''
Rich string.

.. note::

    All options are attributes, that can be modified at all times.

.. note::

    Available methods:

    *   ``A.format()`` :  Formatted string.
    *   ``str(A)`` : Unformatted string.
    *   ``A.isnumeric()`` : Return if the "data" is numeric.
    *   ``int(A)`` : Dummy integer.
    *   ``float(A)`` : Dummy float.

:type data: str, None
:param data: The data.

:type width: None, int
:param width: Print width (formatted print only).

:type color: None, str
:param color: Print color, e.g. "1;32" for bold green (formatted print only).

:type align: ``'<'``, ``'>'``
:param align: Print alignment (formatted print only).

:type dummy: 0, int, float
:param dummy: Dummy numerical value.

:methods:


    '''

    def __init__(self, data, width=None, align='<', color=None, dummy=0):

        self.data  = data
        self.width = width
        self.color = color
        self.align = align
        self.dummy = dummy

    def format(self):
        r'''
Return formatted string: align/width/color are applied.
        '''

        if self.width and self.color:
            fmt = '\x1b[{color:s}m{{0:{align:s}{width:d}.{width:d}s}}\x1b[0m'.format(**self.__dict__)
        elif self.width:
            fmt = '{{0:{align:s}{width:d}.{width:d}s}}'.format(**self.__dict__)
        elif self.color:
            fmt = '\x1b[{color:s}m{{0:{align:s}s}}\x1b[0m'.format(**self.__dict__)
        else:
            fmt = '{{0:{align:s}s}}'.format(**self.__dict__)

        return fmt.format(str(self))

    def isnumeric(self):
        r'''
Return if the "data" is numeric : always zero for this class.
        '''
        return False

    def __str__(self):
        return str(self.data)

    def __int__(self):
        return int(self.dummy)

    def __float__(self):
        return float(self.dummy)

    def __repr__(self):
        return str(self)

    def __lt__(self,other):
        return str(self) < str(other)





from setuptools import setup
from setuptools import find_packages

setup(
    name = 'shelephant',
    license = 'MIT',
    author = 'Tom de Geus',
    author_email = 'tom@geus.me',
    description = 'YAML based shell commands',
    long_description = 'YAML based shell commands',
    keywords = 'YAML, Bash',
    url = 'https://github.com/tdegeus/shelephant',
    packages = find_packages(),
    use_scm_version = {'write_to': 'shelephant/_version.py'},
    setup_requires = ['setuptools_scm'],
    install_requires = ['click', 'pyyaml', 'mergedeep', 'numpy'],
    entry_points = {
        'console_scripts': [
            'shelephant_checksum = shelephant.cli.shelephant_checksum:main',
            'shelephant_dump = shelephant.cli.shelephant_dump:main',
            'shelephant_extract = shelephant.cli.shelephant_extract:main',
            'shelephant_get = shelephant.cli.shelephant_get:main',
            'shelephant_merge = shelephant.cli.shelephant_merge:main',
            'shelephant_parse = shelephant.cli.shelephant_parse:main',
            'shelephant_hostinfo = shelephant.cli.shelephant_hostinfo:main',
            'shelephant_cp = shelephant.cli.shelephant_cp:main',
            'shelephant_mv = shelephant.cli.shelephant_mv:main',
            'shelephant_rm = shelephant.cli.shelephant_rm:main',
            'shelephant_send = shelephant.cli.shelephant_send:main',
        ]})



'''shelephant_dump
    Dump filenames to a YAML-file.

:usage:

    shelephant_dump [options] <file>...

:arguments:

    Files to dump. By default the filenames are written relative to the output file.

:options:

    -o, --output=N
        Output YAML-file. [default: shelephant_dump.yaml]

    -a, --append
        Append existing file.

    -c, --command
        Interpret the input as a command (instead of as filenames).

    --abspath
        Store absolute paths (default: relative to the output file).

    -s, --sort
        Sort filenames.

    -f, --force
        Overwrite output file without prompt.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import os
import subprocess

from .. import version
from .. import yaml


def main():

    try:

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-o', '--output', required=False, default='shelephant_dump.yaml')
        parser.add_argument('-a', '--append', required=False, action='store_true')
        parser.add_argument('-c', '--command', required=False, action='store_true')
        parser.add_argument(      '--abspath', required=False, action='store_true')
        parser.add_argument('-s', '--sort', required=False, action='store_true')
        parser.add_argument('-f', '--force', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('file', nargs='+')
        args = parser.parse_args()

        prefix = os.path.dirname(args.output)
        files = args.file

        if args.command:
            command = ' '.join(files)
            files = sorted(list(filter(None, subprocess.check_output(
                command, shell=True).decode('utf-8').split('\n'))))

        if args.abspath:
            files = [os.path.abspath(file) for file in files]
        else:
            files = [os.path.relpath(file, prefix) for file in files]

        if args.sort:
            files = sorted(files)

        if args.append:

            main = yaml.read(args.output)
            if type(main) != list:
                raise IOError('Can only append a "flat" file')

            files = main + files
            args.force = True

        yaml.dump(args.output, files, args.force)

    except Exception as e:

        print(e)
        return 1


if __name__ == '__main__':

    main()



'''shelephant_extract
    Extract a field from a YAML-file.

    Unless you use --no-path, the function assumes that all data are paths,
    and changes all relative paths from being relative to <input.yaml>
    to being relative to --output.

:usage:

    shelephant_extract [options] <input.yaml> [<key>...]

:arguments:

    <input.yaml>
        The file to read.

    <key>
        The keys to read from the file.

:options:

    -o, --output=N
        Output file. (default: <input.yaml>)

    --no-path
        Do not interpret data as paths.

    -s, --squash
        Squash fields into a single field.

    -f, --force
        Overwrite output file without prompt.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import os
import mergedeep
import functools

from .. import convert
from .. import relpath
from .. import version
from .. import yaml


def main():

    try:

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-o', '--output', required=False)
        parser.add_argument(      '--no-path', required=False, action='store_true')
        parser.add_argument('-s', '--squash', required=False, action='store_true')
        parser.add_argument('-f', '--force', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('input')
        parser.add_argument('key', nargs='*')
        args = parser.parse_args()

        input_dir = os.path.dirname(args.input)
        output = args.output if args.output else args.input
        output_dir = os.path.dirname(output)
        data = {}

        if len(args.key) == 0:
            args.key = ['/']

        for key in args.key:
            key = list(filter(None, key.split('/')))
            files = yaml.read_item(args.input, key)
            if not args.no_path:
                files = relpath.chroot(files, input_dir, output_dir)
            if len(args.key) == 1:
                yaml.dump(output, files, args.force)
                return 0
            container = functools.reduce(lambda x, y: {y: x}, key[:-1], {key[-1]: files})
            mergedeep.merge(data, container)

        if args.squash:
            data = convert.squash(data)

        yaml.dump(output, data, args.force)

    except Exception as e:

        print(e)
        return 1


if __name__ == '__main__':

    main()




import os

def chroot(files, old_root, new_root, in_place=False):
    r'''
Change the root of relative paths.
Skip if all paths are absolute paths.

If ``in_place = True`` the input list is modified 'in place' (and a pointer to it is returned),
otherwise a new list is returned.
    '''

    isabs = [os.path.isabs(file) for file in files]

    if any(isabs) and not all(isabs):
        raise IOError('Specify either relative or absolute files-paths')

    if all(isabs):
        return files

    if not in_place:
        return [os.path.relpath(os.path.abspath(os.path.join(old_root, file)), new_root) for file in files]

    for i in range(len(files)):
        files[i] = os.path.relpath(os.path.abspath(os.path.join(old_root, files[i])), new_root)

    return files


def add_prefix(prefix, files):
    r'''
Add prefix to a list of filenames.
Skip if all paths are absolute paths.

:param str prefix: The prefix.
:param list files: List of paths.
:return: List of paths.
    '''

    isabs = [os.path.isabs(file) for file in files]

    if any(isabs) and not all(isabs):
        raise IOError('Specify either relative or absolute files-paths')

    if all(isabs):
        return files

    return [os.path.normpath(os.path.join(prefix, file)) for file in files]
